import math

import torch
import torch.nn as nn
from torch.autograd import Variable


def setRandomSeed(random_seed=3):
    torch.manual_seed(random_seed)
    torch.cuda.manual_seed(random_seed)
    torch.cuda.manual_seed_all(random_seed)


class AE(nn.Module):
    def __init__(self, N1, N2):
        super(AE, self).__init__()
        self.N1, self.N2 = N1, N2
        self.layer1 = nn.Linear(N1, N2)
        self.layer2 = nn.Linear(N2, N2)
        self.layer3 = nn.Linear(N2, N1)
        self.activation = nn.ReLU()

    def forward(self, inp):
        out1 = self.layer1(inp)
        out1 = self.activation(out1)
        out2 = self.layer2(out1)
        out2 = self.activation(out2)
        out3 = self.layer3(out2)
        return out1, out2, out3


class VAE(nn.Module):
    def __init__(self, N1, N2):
        super(VAE, self).__init__()
        self.N1, self.N2 = N1, N2
        self.layer1 = nn.Linear(N1, N2)
        self.layer2 = nn.Linear(N2, N2)
        self.layer3 = nn.Linear(N2, N1)
        self.activation = nn.ReLU()

    def reparametrize(self, mu, logvar):
        std = logvar.mul(0.5).exp_()
        eps = torch.empty(std.size())
        eps = eps.to(torch.float32).normal_()
        eps = Variable(eps)
        return eps.mul(std).add_(mu)

    def forward(self, inp):
        out1 = self.layer1(inp)
        out1 = self.activation(out1)
        out21 = self.layer2(out1)
        out22 = self.layer2(out1)
        z = self.reparametrize(out21, out22)
        z = self.activation(z)
        out3 = self.layer3(z)
        return out1, out21, out22, z, out3


class SAE(nn.Module):
    def __init__(self, N1, N2):
        super(SAE, self).__init__()
        self.N1, self.N2 = N1, N2
        self.layer1 = nn.Linear(N1, N2)
        self.layer2 = nn.Linear(N2, N2)
        self.layer3 = nn.Linear(N2, N1)
        self.activation = nn.ReLU()

    def forward(self, inp):
        out1 = self.layer1(inp)
        out1 = self.activation(out1)
        out2 = self.layer2(out1)
        out2 = self.activation(out2)
        out3 = self.layer3(out2)
        return out1, out2, out3


class MultiHeadAttention(nn.Module):
    def __init__(self, dim_in, dim_qk, dim_v, num_heads):
        super(MultiHeadAttention, self).__init__()
        assert dim_qk % num_heads == 0 and dim_v % num_heads == 0
        self.dim_in = dim_in
        self.dim_k = dim_qk
        self.dim_q = dim_qk
        self.dim_v = dim_v
        self.num_heads = num_heads
        self.linear_q = nn.Linear(dim_in, dim_qk, bias=False)
        self.linear_k = nn.Linear(dim_in, dim_qk, bias=False)
        self.linear_v = nn.Linear(dim_in, dim_v, bias=False)
        self._norm_fact = 1 / math.sqrt(dim_qk // num_heads)

    def forward(self, input):
        batch, dim_in = input.shape
        assert dim_in == self.dim_in

        dim_k = self.dim_k // self.num_heads
        dim_v = self.dim_v // self.num_heads
        dim_q = self.dim_q // self.num_heads
        q = self.linear_q(input).reshape(batch, self.num_heads, dim_q)
        k = self.linear_k(input).reshape(batch, self.num_heads, dim_k)
        v = self.linear_v(input).reshape(batch, self.num_heads, dim_v)

        dist = torch.matmul(q, k.transpose(1, 2)) * self._norm_fact
        dist = torch.softmax(dist, dim=-1)

        att = torch.matmul(dist, v)
        att = att.transpose(1, 2).reshape(batch, self.dim_v)
        return att


class MA_Discriminator(nn.Module):
    def __init__(self, Net, N1, N2, gradient, num_heads=8):
        super(MA_Discriminator, self).__init__()
        self.N1, self.N2 = N1, N2
        self.model = Net
        self.att = nn.Sequential(MultiHeadAttention(N1, N1 - gradient, N2, num_heads), nn.BatchNorm1d(N2))
        self.layer1 = nn.Linear(N1, N2)
        self.layer2 = nn.Linear(N2, 1)

    def forward(self, input):
        out_list = self.model(input)
        out = out_list[-1]
        out = self.att(out)
        out = self.layer2(out)
        return out


class Discriminator(nn.Module):
    def __init__(self, Net, N1, N2):
        super(Discriminator, self).__init__()
        self.N1, self.N2 = N1, N2
        self.model = Net
        self.layer1 = nn.Linear(N1, N2)
        self.layer2 = nn.Linear(N2, 1)

    def forward(self, input):
        out_list = self.model(input)
        out = out_list[-1]
        out = self.layer1(out)
        out = self.layer2(out)
        return out
