import warnings

import numpy as np
import torch
import torch.nn as nn

warnings.filterwarnings('ignore')
np.set_printoptions(precision=3)


class SAE:
    def __init__(self, epochs=50, batch_size=10, lr=1e-2):
        self.epochs = epochs
        self.batch_size = batch_size
        self.lr = lr
        self.loss_func = nn.MSELoss()

    def sparse_loss(self, input, hidden, output):
        def kl(p, q):
            q = torch.nn.functional.softmax(q, dim=0)
            q = torch.sum(q,
                          dim=0) / self.batch_size
            s1 = torch.sum(p * torch.log(p / q))
            return s1

        criterion = nn.MSELoss()
        MSE_loss = criterion(input, output)
        rho_hat = torch.sum(hidden, dim=0, keepdim=True)
        sparsity_penalty = 3 * kl(0.01, rho_hat)
        return MSE_loss + sparsity_penalty

    def pre_train(self, dataLoader, net, optimizer):
        for index, (order, label, data) in enumerate(dataLoader):
            net.train()
            out1, out2, out3 = net(data)
            loss = self.sparse_loss(data, out2, out3)
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
        return net, optimizer

    def train(self, dataLoader, data_size, net, optimizer):
        train_loss = 0
        hid1, hid2, out, order_list = [], [], [], []
        for index, (order, label, data) in enumerate(dataLoader):
            net.train()
            out1, out2, out3 = net(data)
            loss = self.loss_func(out3, data)
            train_loss += np.abs(loss.item()) * data.shape[0]
            loss = self.sparse_loss(data, out2, out3)
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
            hid1.append(out1)
            hid2.append(out2)
            out.append(out3)
            order_list.append(order)
        return train_loss / data_size, net, optimizer, hid1, hid2, out, order_list

    def test(self, dataLoader, data_size, net):
        test_loss = 0
        hid1, hid2, out, order_list = [], [], [], []
        for index, (order, label, data) in enumerate(dataLoader):
            net.eval()
            out1, out2, out3 = net(data)
            loss = self.loss_func(out3, data)
            test_loss += np.abs(loss.item()) * data.shape[0]
            hid1.append(out1)
            hid2.append(out2)
            out.append(out3)
            order_list.append(order)
        return test_loss / data_size, hid1, hid2, out, order_list
